@@ -1,13 +1,10 @@
 """PGSync Base class."""
 import logging
 import os
-import sys
-import warnings
 from collections import defaultdict
-from typing import Dict, List, Optional, Tuple
+from typing import Dict, List, Optional
 
 import sqlalchemy as sa
-import sqlparse
 from sqlalchemy.dialects import postgresql  # noqa
 from sqlalchemy.orm import sessionmaker
 
@@ -29,15 +26,11 @@ from .exc import (
     TableNotFoundError,
 )
 from .node import Node
-from .settings import (
-    PG_SSLMODE,
-    PG_SSLROOTCERT,
-    QUERY_CHUNK_SIZE,
-    QUERY_LITERAL_BINDS,
-)
+from .settings import PG_SSLMODE, PG_SSLROOTCERT, QUERY_CHUNK_SIZE
 from .trigger import CREATE_TRIGGER_TEMPLATE
 from .urls import get_postgres_url
-from .view import create_view, drop_view
+from .utils import compiled_query
+from .view import create_view, DropView, is_view, RefreshView
 
 try:
     import citext  # noqa
@@ -75,13 +68,17 @@ class TupleIdentifierType(sa.types.UserDefinedType):
 class Base(object):
     def __init__(self, database: str, verbose: bool = False, *args, **kwargs):
         """Initialize the base class constructor."""
-        self.__engine = pg_engine(database, **kwargs)
+        self.__engine: sa.engine.Engine = pg_engine(database, **kwargs)
         self.__schemas: Optional[dict] = None
         # models is a dict of f'{schema}.{table}'
         self.models: Dict[str] = {}
-        self.__metadata: dict = {}
-        self.__indices: dict = {}
+        self.__metadata: Dict[str] = {}
+        self.__indices: Dict[str] = {}
+        self.__views: Dict[str] = {}
+        self.__materialized_views: Dict[str] = {}
+        self.__tables: Dict[str] = {}
         self.verbose: bool = verbose
+        self._conn = None
 
     def connect(self) -> None:
         """Connect to database."""
@@ -112,10 +109,11 @@ class Base(object):
                 f"Invalid user permission {permissions}"
             )
 
-        # Azure usernames are of the form username@host on the SQLAlchemy
-        # engine - engine.url.username@host but stored in
-        # pg_user as just user.
-        # we need to extract the real username from: username@host.
+        # Microsoft Azure usernames are of the form username@host on
+        # the SQLAlchemy engine.
+        # e.g engine.url.username@host
+        # but stored in pg_user as just user.
+        # we need to extract the real username from: username@host
         host_part: str = self.engine.url.host.split(".")[0]
         username: str = username.split(f"@{host_part}")[0]
 
@@ -130,7 +128,10 @@ class Base(object):
                                 sa.column("usename") == username,
                                 sa.or_(
                                     *[
-                                        (sa.column(permission) == True)
+                                        (
+                                            sa.column(permission)
+                                            == True  # noqa E712
+                                        )
                                         for permission in permissions
                                     ]
                                 ),
@@ -144,7 +145,7 @@ class Base(object):
             )
 
     # Tables...
-    def model(self, table: str, schema: str) -> dict:
+    def model(self, table: str, schema: str) -> sa.sql.Alias:
         """Get an SQLAlchemy model representation from a table.
 
         Args:
@@ -185,6 +186,10 @@ class Base(object):
 
         return self.models[name]
 
+    @property
+    def conn(self):
+        return self._conn
+
     @property
     def database(self) -> str:
         """str: Get the database name."""
@@ -196,7 +201,7 @@ class Base(object):
         return Session()
 
     @property
-    def engine(self):
+    def engine(self) -> sa.engine.Engine:
         """Get the database engine."""
         return self.__engine
 
@@ -210,31 +215,43 @@ class Base(object):
                     self.__schemas.remove(schema)
         return self.__schemas
 
-    def indices(self, table: str) -> list:
+    def views(self, schema: str) -> list:
+        """Get all materialized and non-materialized views."""
+        return self._views(schema) + self._materialized_views(schema)
+
+    def _views(self, schema: str) -> list:
+        """Get all non-materialized views."""
+        if schema not in self.__views:
+            self.__views[schema] = []
+            for table in sa.inspect(self.engine).get_view_names(schema):
+                if is_view(self.engine, schema, table, materialized=False):
+                    self.__views[schema].append(table)
+        return self.__views[schema]
+
+    def _materialized_views(self, schema: str) -> list:
+        """Get all materialized views."""
+        if schema not in self.__materialized_views:
+            self.__materialized_views[schema] = []
+            for table in sa.inspect(self.engine).get_view_names(schema):
+                if is_view(self.engine, schema, table, materialized=True):
+                    self.__materialized_views[schema].append(table)
+        return self.__materialized_views[schema]
+
+    def indices(self, table: str, schema: str) -> list:
         """Get the database table indexes."""
         if table not in self.__indices:
-            self.__indices[table] = sa.inspect(self.engine).get_indexes(table)
+            self.__indices[table] = sa.inspect(self.engine).get_indexes(
+                table, schema=schema
+            )
         return self.__indices[table]
 
-    def tables(self, schema: str) -> List:
-        """:obj:`list` of :obj:`str`: Get all tables.
-
-        returns the fully qualified table name with schema {schema}.{table}
-        """
-        if schema not in self.__metadata:
-            metadata = sa.MetaData(schema=schema)
-            metadata.reflect(self.__engine)
-            self.__metadata[schema] = metadata
-        metadata = self.__metadata[schema]
-        return metadata.tables.keys()
-
-    def _get_schema(self, schema: str, table: str) -> Tuple[str, str]:
-        pairs: list = table.split(".")
-        if len(pairs) == 2:
-            return pairs[0], pairs[1]
-        if len(pairs) == 1:
-            return schema, pairs[0]
-        raise ValueError(f"Invalid definition {table} for schema: {schema}")
+    def tables(self, schema: str) -> list:
+        """Get the table names for current schema."""
+        if schema not in self.__tables:
+            self.__tables[schema] = sa.inspect(self.engine).get_table_names(
+                schema
+            )
+        return self.__tables[schema]
 
     def truncate_table(self, table: str, schema: str = DEFAULT_SCHEMA) -> None:
         """Truncate a table.
@@ -248,9 +265,6 @@ class Base(object):
             schema (str): The database schema
 
         """
-        if not table.startswith(f"{schema}."):
-            table = f"{schema}.{table}"
-        schema, table = self._get_schema(schema, table)
         logger.debug(f"Truncating table: {schema}.{table}")
         query = f'TRUNCATE TABLE "{schema}"."{table}" CASCADE'
         self.execute(query)
@@ -342,9 +356,11 @@ class Base(object):
         txmax: Optional[int] = None,
         upto_lsn: Optional[int] = None,
         upto_nchanges: Optional[int] = None,
-    ):
+        limit: Optional[int] = None,
+        offset: Optional[int] = None,
+    ) -> sa.sql.Select:
         filters: list = []
-        statement: str = sa.select(
+        statement: sa.sql.Select = sa.select(
             [sa.column("xid"), sa.column("data")]
         ).select_from(
             func(
@@ -353,7 +369,7 @@ class Base(object):
                 upto_nchanges,
             )
         )
-        if txmin:
+        if txmin is not None:
             filters.append(
                 sa.cast(
                     sa.cast(sa.column("xid"), sa.Text),
@@ -361,7 +377,7 @@ class Base(object):
                 )
                 >= txmin
             )
-        if txmax:
+        if txmax is not None:
             filters.append(
                 sa.cast(
                     sa.cast(sa.column("xid"), sa.Text),
@@ -371,7 +387,11 @@ class Base(object):
             )
         if filters:
             statement = statement.where(sa.and_(*filters))
-        return self.fetchall(statement)
+        if limit is not None:
+            statement = statement.limit(limit)
+        if offset is not None:
+            statement = statement.offset(offset)
+        return statement
 
     def logical_slot_get_changes(
         self,
@@ -380,7 +400,9 @@ class Base(object):
         txmax: Optional[int] = None,
         upto_lsn: Optional[int] = None,
         upto_nchanges: Optional[int] = None,
-    ):
+        limit: Optional[int] = None,
+        offset: Optional[int] = None,
+    ) -> List[sa.engine.row.LegacyRow]:
         """Get/Consume changes from a logical replication slot.
 
         To get one change and data in existing replication slot:
@@ -389,14 +411,17 @@ class Base(object):
         To get ALL changes and data in existing replication slot:
         SELECT * FROM PG_LOGICAL_SLOT_GET_CHANGES('testdb', NULL, NULL)
         """
-        return self._logical_slot_changes(
+        statement: sa.sql.Select = self._logical_slot_changes(
             slot_name,
             sa.func.PG_LOGICAL_SLOT_GET_CHANGES,
             txmin=txmin,
             txmax=txmax,
             upto_lsn=upto_lsn,
             upto_nchanges=upto_nchanges,
+            limit=limit,
+            offset=offset,
         )
+        return self.fetchall(statement)
 
     def logical_slot_peek_changes(
         self,
@@ -405,12 +430,34 @@ class Base(object):
         txmax: Optional[int] = None,
         upto_lsn: Optional[int] = None,
         upto_nchanges: Optional[int] = None,
-    ):
+        limit: Optional[int] = None,
+        offset: Optional[int] = None,
+    ) -> List[sa.engine.row.LegacyRow]:
         """Peek a logical replication slot without consuming changes.
 
         SELECT * FROM PG_LOGICAL_SLOT_PEEK_CHANGES('testdb', NULL, 1)
         """
-        return self._logical_slot_changes(
+        statement: sa.sql.Select = self._logical_slot_changes(
+            slot_name,
+            sa.func.PG_LOGICAL_SLOT_PEEK_CHANGES,
+            txmin=txmin,
+            txmax=txmax,
+            upto_lsn=upto_lsn,
+            upto_nchanges=upto_nchanges,
+            limit=limit,
+            offset=offset,
+        )
+        return self.fetchall(statement)
+
+    def logical_slot_count_changes(
+        self,
+        slot_name: str,
+        txmin: Optional[int] = None,
+        txmax: Optional[int] = None,
+        upto_lsn: Optional[int] = None,
+        upto_nchanges: Optional[int] = None,
+    ) -> int:
+        statement: sa.sql.Select = self._logical_slot_changes(
             slot_name,
             sa.func.PG_LOGICAL_SLOT_PEEK_CHANGES,
             txmin=txmin,
@@ -418,217 +465,115 @@ class Base(object):
             upto_lsn=upto_lsn,
             upto_nchanges=upto_nchanges,
         )
+        with self.__engine.connect() as conn:
+            return conn.execute(
+                statement.with_only_columns([sa.func.COUNT()])
+            ).scalar()
 
     # Views...
-    def _primary_keys(
-        self, schema: str, tables: List[str]
-    ) -> sa.sql.selectable.Select:
-        with warnings.catch_warnings():
-            warnings.simplefilter("ignore", category=sa.exc.SAWarning)
-            pg_class = self.model("pg_class", "pg_catalog")
-            pg_index = self.model("pg_index", "pg_catalog")
-            pg_attribute = self.model("pg_attribute", "pg_catalog")
-            pg_namespace = self.model("pg_namespace", "pg_catalog")
-
-        alias = pg_class.alias("x")
-        inclause: list = []
-        for table in tables:
-            pairs = table.split(".")
-            if len(pairs) == 1:
-                inclause.append(
-                    self.__engine.dialect.identifier_preparer.quote(pairs[0])
-                )
-            elif len(pairs) == 2:
-                inclause.append(
-                    f"{pairs[0]}.{self.__engine.dialect.identifier_preparer.quote(pairs[-1])}"
-                )
-            else:
-                raise Exception(
-                    f"cannot determine schema and table from {table}"
-                )
-
-        return (
-            sa.select(
-                [
-                    sa.func.REPLACE(
-                        sa.func.REVERSE(
-                            sa.func.SPLIT_PART(
-                                sa.func.REVERSE(
-                                    sa.cast(
-                                        sa.cast(
-                                            pg_index.c.indrelid,
-                                            sa.dialects.postgresql.REGCLASS,
-                                        ),
-                                        sa.Text,
-                                    )
-                                ),
-                                ".",
-                                1,
-                            )
-                        ),
-                        '"',
-                        "",
-                    ).label("table_name"),
-                    sa.func.ARRAY_AGG(pg_attribute.c.attname).label(
-                        "primary_keys"
-                    ),
-                ]
-            )
-            .join(
-                pg_attribute,
-                pg_attribute.c.attrelid == pg_index.c.indrelid,
-            )
-            .join(
-                pg_class,
-                pg_class.c.oid == pg_index.c.indexrelid,
-            )
-            .join(
-                alias,
-                alias.c.oid == pg_index.c.indrelid,
-            )
-            .join(
-                pg_namespace,
-                pg_namespace.c.oid == pg_class.c.relnamespace,
-            )
-            .where(
-                *[
-                    pg_namespace.c.nspname.notin_(["pg_catalog", "pg_toast"]),
-                    pg_index.c.indisprimary,
-                    sa.cast(
-                        sa.cast(
-                            pg_index.c.indrelid,
-                            sa.dialects.postgresql.REGCLASS,
-                        ),
-                        sa.Text,
-                    ).in_(inclause),
-                    pg_attribute.c.attnum == sa.any_(pg_index.c.indkey),
-                ]
-            )
-            .group_by(pg_index.c.indrelid)
-        )
-
-    def _foreign_keys(
-        self, schema: str, tables: List[str]
-    ) -> sa.sql.selectable.Select:
-        with warnings.catch_warnings():
-            warnings.simplefilter("ignore", category=sa.exc.SAWarning)
-            table_constraints = self.model(
-                "table_constraints",
-                "information_schema",
-            )
-            key_column_usage = self.model(
-                "key_column_usage",
-                "information_schema",
-            )
-            constraint_column_usage = self.model(
-                "constraint_column_usage",
-                "information_schema",
-            )
-
-        return (
-            sa.select(
-                [
-                    table_constraints.c.table_name,
-                    sa.func.ARRAY_AGG(
-                        sa.cast(
-                            key_column_usage.c.column_name,
-                            sa.TEXT,
-                        )
-                    ).label("foreign_keys"),
-                ]
-            )
-            .join(
-                key_column_usage,
-                sa.and_(
-                    key_column_usage.c.constraint_name
-                    == table_constraints.c.constraint_name,
-                    key_column_usage.c.table_schema
-                    == table_constraints.c.table_schema,
-                    key_column_usage.c.table_schema == schema,
-                ),
-            )
-            .join(
-                constraint_column_usage,
-                sa.and_(
-                    constraint_column_usage.c.constraint_name
-                    == table_constraints.c.constraint_name,
-                    constraint_column_usage.c.table_schema
-                    == table_constraints.c.table_schema,
-                ),
-            )
-            .where(
-                *[
-                    table_constraints.c.table_name.in_(tables),
-                    table_constraints.c.constraint_type == "FOREIGN KEY",
-                ]
-            )
-            .group_by(table_constraints.c.table_name)
-        )
-
     def create_view(
         self, schema: str, tables: list, user_defined_fkey_tables: dict
-    ):
+    ) -> None:
         create_view(
-            self.engine, schema, tables, user_defined_fkey_tables, self
+            self.engine,
+            self.model,
+            self.fetchall,
+            schema,
+            tables,
+            user_defined_fkey_tables,
+            self._materialized_views(schema),
         )
 
-    def drop_view(self, schema: str):
-        drop_view(self.engine, schema)
+    def drop_view(self, schema: str) -> None:
+        """Drop a view."""
+        logger.debug(f"Dropping view: {schema}.{MATERIALIZED_VIEW}")
+        self.engine.execute(DropView(schema, MATERIALIZED_VIEW))
+        logger.debug(f"Dropped view: {schema}.{MATERIALIZED_VIEW}")
+
+    def refresh_view(
+        self, name: str, schema: str, concurrently: bool = False
+    ) -> None:
+        """Refresh a materialized view."""
+        logger.debug(f"Refreshing view: {schema}.{name}")
+        self.engine.execute(
+            RefreshView(schema, name, concurrently=concurrently)
+        )
+        logger.debug(f"Refreshed view: {schema}.{name}")
 
     # Triggers...
     def create_triggers(
-        self, schema: str, tables: Optional[List[str]] = None
+        self,
+        schema: str,
+        tables: Optional[List[str]] = None,
+        join_queries: bool = False,
     ) -> None:
         """Create a database triggers."""
-        self.execute(
-            CREATE_TRIGGER_TEMPLATE.replace(
-                MATERIALIZED_VIEW, f"{schema}.{MATERIALIZED_VIEW}"
-            )
-        )
-        views = sa.inspect(self.engine).get_view_names(schema)
-        queries = []
+        queries: List[str] = []
         for table in self.tables(schema):
-            schema, table = self._get_schema(schema, table)
-            if (tables and table not in tables) or (table in views):
+            if (tables and table not in tables) or (
+                table in self.views(schema)
+            ):
                 continue
             logger.debug(f"Creating trigger on table: {schema}.{table}")
-            for name, for_each, tg_op in [
+            for name, level, tg_op in [
                 ("notify", "ROW", ["INSERT", "UPDATE", "DELETE"]),
                 ("truncate", "STATEMENT", ["TRUNCATE"]),
             ]:
                 self.drop_triggers(schema, [table])
                 queries.append(
-                    sa.DDL(
-                        f'CREATE TRIGGER "{table}_{name}" '
-                        f'AFTER {" OR ".join(tg_op)} ON "{schema}"."{table}" '
-                        f"FOR EACH {for_each} EXECUTE PROCEDURE "
-                        f"{TRIGGER_FUNC}()",
-                    )
+                    f'CREATE TRIGGER "{table}_{name}" '
+                    f'AFTER {" OR ".join(tg_op)} ON "{schema}"."{table}" '
+                    f"FOR EACH {level} EXECUTE PROCEDURE "
+                    f"{schema}.{TRIGGER_FUNC}()",
                 )
-
-        for query in queries:
-            self.execute(query)
+        if join_queries:
+            if queries:
+                self.execute("; ".join(queries))
+        else:
+            for query in queries:
+                self.execute(sa.DDL(query))
 
     def drop_triggers(
-        self, schema: str, tables: Optional[List[str]] = None
+        self,
+        schema: str,
+        tables: Optional[List[str]] = None,
+        join_queries: bool = False,
     ) -> None:
         """Drop all pgsync defined triggers in database."""
+        queries: List[str] = []
         for table in self.tables(schema):
-            schema, table = self._get_schema(schema, table)
             if tables and table not in tables:
                 continue
             logger.debug(f"Dropping trigger on table: {schema}.{table}")
             for name in ("notify", "truncate"):
-                query = (
+                queries.append(
                     f'DROP TRIGGER IF EXISTS "{table}_{name}" ON '
                     f'"{schema}"."{table}"'
                 )
-                self.execute(query)
+        if join_queries:
+            if queries:
+                self.execute("; ".join(queries))
+        else:
+            for query in queries:
+                self.execute(sa.DDL(query))
+
+    def create_function(self, schema: str) -> None:
+        self.execute(
+            CREATE_TRIGGER_TEMPLATE.replace(
+                MATERIALIZED_VIEW,
+                f"{schema}.{MATERIALIZED_VIEW}",
+            ).replace(
+                TRIGGER_FUNC,
+                f"{schema}.{TRIGGER_FUNC}",
+            )
+        )
+
+    def drop_function(self, schema: str) -> None:
+        self.execute(f'DROP FUNCTION IF EXISTS "{schema}".{TRIGGER_FUNC}()')
 
     def disable_triggers(self, schema: str) -> None:
         """Disable all pgsync defined triggers in database."""
         for table in self.tables(schema):
-            schema, table = self._get_schema(schema, table)
             logger.debug(f"Disabling trigger on table: {schema}.{table}")
             for name in ("notify", "truncate"):
                 query = (
@@ -640,7 +585,6 @@ class Base(object):
     def enable_triggers(self, schema: str) -> None:
         """Enable all pgsync defined triggers in database."""
         for table in self.tables(schema):
-            schema, table = self._get_schema(schema, table)
             logger.debug(f"Enabling trigger on table: {schema}.{table}")
             for name in ("notify", "truncate"):
                 query = (
@@ -650,7 +594,7 @@ class Base(object):
                 self.execute(query)
 
     @property
-    def txid_current(self):
+    def txid_current(self) -> int:
         """
         Get last committed transaction id from the database.
 
@@ -715,7 +659,7 @@ class Base(object):
                 raise
         return value
 
-    def parse_logical_slot(self, row):
+    def parse_logical_slot(self, row: str):
         def _parse_logical_slot(data):
 
             while True:
@@ -736,7 +680,9 @@ class Base(object):
                 data = f"{data[match.span()[1]:]} "
                 yield key, value
 
-        payload = dict(schema=None, tg_op=None, table=None, old={}, new={})
+        payload: dict = dict(
+            schema=None, tg_op=None, table=None, old={}, new={}
+        )
 
         match = LOGICAL_SLOT_PREFIX.search(row)
         if not match:
@@ -744,7 +690,7 @@ class Base(object):
 
         payload.update(match.groupdict())
         span = match.span()
-        # trailing space is deliberate
+        # including trailing space below is deliberate
         suffix: str = f"{row[span[1]:]} "
         tg_op: str = payload["tg_op"]
 
@@ -778,8 +724,12 @@ class Base(object):
         return payload
 
     # Querying...
-
-    def execute(self, statement, values=None, options=None):
+    def execute(
+        self,
+        statement: sa.sql.Select,
+        values: dict = None,
+        options: dict = None,
+    ) -> None:
         """Execute a query statement."""
         conn = self.__engine.connect()
         try:
@@ -791,7 +741,9 @@ class Base(object):
             logger.exception(f"Exception {e}")
             raise
 
-    def fetchone(self, statement, label=None, literal_binds=False):
+    def fetchone(
+        self, statement: sa.sql.Select, label=None, literal_binds=False
+    ) -> sa.engine.Row:
         """Fetch one row query."""
         if self.verbose:
             compiled_query(statement, label=label, literal_binds=literal_binds)
@@ -805,7 +757,12 @@ class Base(object):
             raise
         return row
 
-    def fetchall(self, statement, label=None, literal_binds=False):
+    def fetchall(
+        self,
+        statement: sa.sql.Select,
+        label: Optional[str] = None,
+        literal_binds: bool = False,
+    ) -> List[sa.engine.Row]:
         """Fetch all rows from a query statement."""
         if self.verbose:
             compiled_query(statement, label=label, literal_binds=literal_binds)
@@ -821,7 +778,7 @@ class Base(object):
 
     def fetchmany(
         self,
-        statement: sa.sql.selectable.Select,
+        statement: sa.sql.Select,
         chunk_size: Optional[int] = None,
     ):
         chunk_size: int = chunk_size or QUERY_CHUNK_SIZE
@@ -833,7 +790,7 @@ class Base(object):
                 for keys, row, *primary_keys in partition:
                     yield keys, row, primary_keys
 
-    def fetchcount(self, statement):
+    def fetchcount(self, statement: sa.sql.Subquery) -> int:
         with self.__engine.connect() as conn:
             return conn.execute(
                 statement.original.with_only_columns(
@@ -935,16 +892,16 @@ def get_foreign_keys(node_a: Node, node_b: Node) -> dict:
 
 
 def pg_engine(
-    database,
-    user=None,
-    host=None,
-    password=None,
-    port=None,
-    echo=False,
-    sslmode=None,
-    sslrootcert=None,
-):
-    connect_args = {}
+    database: str,
+    user: Optional[str] = None,
+    host: Optional[str] = None,
+    password: Optional[str] = None,
+    port: Optional[str] = None,
+    echo: bool = False,
+    sslmode: Optional[str] = None,
+    sslrootcert: Optional[str] = None,
+) -> sa.engine.Engine:
+    connect_args: dict = {}
     sslmode: str = sslmode or PG_SSLMODE
     sslrootcert: str = sslrootcert or PG_SSLROOTCERT
 
@@ -980,8 +937,8 @@ def pg_engine(
 
 
 def pg_execute(
-    engine,
-    query,
+    engine: sa.engine.Engine,
+    query: str,
     values: Optional[list] = None,
     options: Optional[dict] = None,
 ) -> None:
@@ -997,7 +954,7 @@ def pg_execute(
         raise
 
 
-def create_schema(engine, schema) -> None:
+def create_schema(engine: sa.engine.Engine, schema: str) -> None:
     """Create database schema."""
     if schema != DEFAULT_SCHEMA:
         engine.execute(sa.schema.CreateSchema(schema))
@@ -1006,7 +963,7 @@ def create_schema(engine, schema) -> None:
 def create_database(database: str, echo: bool = False) -> None:
     """Create a database."""
     logger.debug(f"Creating database: {database}")
-    engine = pg_engine(database="postgres", echo=echo)
+    engine: sa.engine.Engine = pg_engine(database="postgres", echo=echo)
     pg_execute(engine, f'CREATE DATABASE "{database}"')
     logger.debug(f"Created database: {database}")
 
@@ -1014,7 +971,7 @@ def create_database(database: str, echo: bool = False) -> None:
 def drop_database(database: str, echo: bool = False) -> None:
     """Drop a database."""
     logger.debug(f"Dropping database: {database}")
-    engine = pg_engine(database="postgres", echo=echo)
+    engine: sa.engine.Engine = pg_engine(database="postgres", echo=echo)
     pg_execute(engine, f'DROP DATABASE IF EXISTS "{database}"')
     logger.debug(f"Dropped database: {database}")
 
@@ -1024,7 +981,7 @@ def create_extension(
 ) -> None:
     """Create a database extension."""
     logger.debug(f"Creating extension: {extension}")
-    engine = pg_engine(database=database, echo=echo)
+    engine: sa.engine.Engine = pg_engine(database=database, echo=echo)
     pg_execute(engine, f'CREATE EXTENSION IF NOT EXISTS "{extension}"')
     logger.debug(f"Created extension: {extension}")
 
@@ -1032,32 +989,6 @@ def create_extension(
 def drop_extension(database: str, extension: str, echo: bool = False) -> None:
     """Drop a database extension."""
     logger.debug(f"Dropping extension: {extension}")
-    engine = pg_engine(database=database, echo=echo)
+    engine: sa.engine.Engine = pg_engine(database=database, echo=echo)
     pg_execute(engine, f'DROP EXTENSION IF EXISTS "{extension}"')
     logger.debug(f"Dropped extension: {extension}")
-
-
-def compiled_query(
-    query: str, label: Optional[str] = None, literal_binds: bool = False
-) -> None:
-    """Compile an SQLAlchemy query with an optional label."""
-
-    # overide env value of literal_binds
-    if QUERY_LITERAL_BINDS:
-        literal_binds = QUERY_LITERAL_BINDS
-
-    query: str = str(
-        query.compile(
-            dialect=sa.dialects.postgresql.dialect(),
-            compile_kwargs={"literal_binds": literal_binds},
-        )
-    )
-    query: str = sqlparse.format(query, reindent=True, keyword_case="upper")
-    if label:
-        logger.debug(f"\033[4m{label}:\033[0m\n{query}")
-        sys.stdout.write(f"\033[4m{label}:\033[0m\n{query}\n")
-    else:
-        logging.debug(f"{query}")
-        sys.stdout.write(f"{query}\n")
-    sys.stdout.write("-" * 79)
-    sys.stdout.write("\n")
