@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 
 """Sync module."""
+import asyncio
 import json
 import logging
 import os
@@ -14,12 +15,13 @@ from typing import AnyStr, Generator, List, Optional, Set
 
 import click
 import sqlalchemy as sa
+import sqlparse
 from psycopg2 import OperationalError
 from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
 from sqlalchemy.sql import Values
 
 from . import __version__
-from .base import Base, compiled_query, get_foreign_keys, TupleIdentifierType
+from .base import Base, get_foreign_keys, TupleIdentifierType
 from .constants import (
     DELETE,
     INSERT,
@@ -43,15 +45,25 @@ from .querybuilder import QueryBuilder
 from .redisqueue import RedisQueue
 from .settings import (
     CHECKPOINT_PATH,
+    JOIN_QUERIES,
     LOG_INTERVAL,
+    LOGICAL_SLOT_CHUNK_SIZE,
     NTHREADS_POLLDB,
     POLL_TIMEOUT,
     REDIS_POLL_INTERVAL,
     REDIS_WRITE_CHUNK_SIZE,
     REPLICATION_SLOT_CLEANUP_INTERVAL,
+    USE_ASYNC,
 )
 from .transform import get_private_keys, transform
-from .utils import exit_handler, get_config, show_settings, threaded, Timer
+from .utils import (
+    compiled_query,
+    exception,
+    get_config,
+    show_settings,
+    threaded,
+    Timer,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -62,9 +74,9 @@ class Sync(Base):
     def __init__(
         self,
         document: dict,
-        verbose: Optional[bool] = False,
-        validate: Optional[bool] = True,
-        repl_slots: Optional[bool] = True,
+        verbose: bool = False,
+        validate: bool = True,
+        repl_slots: bool = True,
         **kwargs,
     ):
         """Constructor."""
@@ -90,6 +102,7 @@ class Sync(Base):
         )
         self.redis: RedisQueue = RedisQueue(self.__name)
         self.tree: Tree = Tree(self)
+        self.__root: Node = None
         if validate:
             self.validate(repl_slots=repl_slots)
             self.create_setting()
@@ -98,7 +111,13 @@ class Sync(Base):
         )
         self.count: dict = dict(xlog=0, db=0, redis=0)
 
-    def validate(self, repl_slots: Optional[bool] = True) -> None:
+    @property
+    def root(self) -> str:
+        if self.__root is None:
+            self.__root = self.tree.build(self.nodes)
+        return self.__root
+
+    def validate(self, repl_slots: bool = True) -> None:
         """Perform all validation right away."""
 
         # ensure v2 compatible schema
@@ -172,14 +191,18 @@ class Sync(Base):
                 f"read/writable"
             )
 
-        root: Node = self.tree.build(self.nodes)
-        root.display()
-        for node in root.traverse_breadth_first():
-            pass
+        self.root.display()
+        for node in self.root.traverse_breadth_first():
+            # ensure all base tables have at least one primary_key
+            for table in node.base_tables:
+                model: sa.sql.Alias = self.model(table, node.schema)
+                if not model.primary_keys:
+                    raise PrimaryKeyNotFoundError(
+                        f"No primary key(s) for base table: {table}"
+                    )
 
     def analyze(self) -> None:
-        root: Node = self.tree.build(self.nodes)
-        for node in root.traverse_breadth_first():
+        for node in self.root.traverse_breadth_first():
 
             if node.is_root:
                 continue
@@ -205,7 +228,7 @@ class Sync(Base):
                     node,
                 )
 
-            for index in self.indices(node.table):
+            for index in self.indices(node.table, node.schema):
                 columns: list = foreign_keys.get(node.name, [])
                 if set(columns).issubset(index.get("column_names", [])) or set(
                     columns
@@ -221,23 +244,22 @@ class Sync(Base):
                     f'Missing index on table "{node.table}" for columns: '
                     f"{columns}\n"
                 )
-                command: str = (
+                query: str = sqlparse.format(
                     f'CREATE INDEX idx_{node.table}_{"_".join(columns)} ON '
-                    f'{node.table} ({", ".join(columns)})'
-                )
-                sys.stdout.write(
-                    f'Create one with: "\033[4m{command}\033[0m"\n'
+                    f'{node.table} ({", ".join(columns)})',
+                    reindent=True,
+                    keyword_case="upper",
                 )
+                sys.stdout.write(f'Create one with: "\033[4m{query}\033[0m"\n')
                 sys.stdout.write("-" * 80)
                 sys.stdout.write("\n")
                 sys.stdout.flush()
 
     def create_setting(self) -> None:
         """Create Elasticsearch setting and mapping if required."""
-        root: Node = self.tree.build(self.nodes)
         self.es._create_setting(
             self.index,
-            root,
+            self.root,
             setting=self.setting,
             mapping=self.mapping,
             routing=self.routing,
@@ -245,19 +267,25 @@ class Sync(Base):
 
     def setup(self) -> None:
         """Create the database triggers and replication slot."""
+
+        join_queries: bool = JOIN_QUERIES
+
         self.teardown(drop_view=False)
 
         for schema in self.schemas:
-            tables: Set = set([])
+            self.create_function(schema)
+            tables: Set = set()
             # tables with user defined foreign keys
             user_defined_fkey_tables: dict = {}
 
-            root: Node = self.tree.build(self.nodes)
-            for node in root.traverse_breadth_first():
+            for node in self.root.traverse_breadth_first():
                 if node.schema != schema:
                     continue
                 tables |= set(node.relationship.through_tables)
                 tables |= set([node.table])
+                # we also need to bootstrap the base tables
+                tables |= set(node.base_tables)
+
                 # we want to get both the parent and the child keys here
                 # even though only one of them is the foreign_key.
                 # this is because we define both in the schema but
@@ -268,15 +296,20 @@ class Sync(Base):
                 if node.relationship.foreign_key.child:
                     columns.extend(node.relationship.foreign_key.child)
                 if columns:
-                    user_defined_fkey_tables.setdefault(node.table, set([]))
+                    user_defined_fkey_tables.setdefault(node.table, set())
                     user_defined_fkey_tables[node.table] |= set(columns)
             if tables:
                 self.create_view(schema, tables, user_defined_fkey_tables)
-                self.create_triggers(schema, tables=tables)
+                self.create_triggers(
+                    schema, tables=tables, join_queries=join_queries
+                )
         self.create_replication_slot(self.__name)
 
     def teardown(self, drop_view: bool = True) -> None:
         """Drop the database triggers and replication slot."""
+
+        join_queries: bool = JOIN_QUERIES
+
         try:
             os.unlink(self._checkpoint_file)
         except OSError:
@@ -285,21 +318,27 @@ class Sync(Base):
         self.redis.delete()
 
         for schema in self.schemas:
-            tables: Set = set([])
-            root: Node = self.tree.build(self.nodes)
-            for node in root.traverse_breadth_first():
+            tables: Set = set()
+            for node in self.root.traverse_breadth_first():
                 tables |= set(node.relationship.through_tables)
                 tables |= set([node.table])
-            self.drop_triggers(schema=schema, tables=tables)
+                # we also need to teardown the base tables
+                tables |= set(node.base_tables)
+            self.drop_triggers(
+                schema=schema, tables=tables, join_queries=join_queries
+            )
             if drop_view:
-                self.drop_view(schema=schema)
+                self.drop_view(schema)
+
+            self.drop_function(schema)
+
         self.drop_replication_slot(self.__name)
 
-    def get_doc_id(self, primary_keys: List[str]) -> str:
+    def get_doc_id(self, primary_keys: List[str], table: str) -> str:
         """Get the Elasticsearch document id from the primary keys."""
         if not primary_keys:
             raise PrimaryKeyNotFoundError(
-                "No primary key found on target table"
+                f"No primary key found on table: {table}"
             )
         return f"{PRIMARY_KEY_DELIMITER}".join(map(str, primary_keys))
 
@@ -329,65 +368,81 @@ class Sync(Base):
         TODO: We can also process all INSERTS together and rearrange
         them as done below
         """
-        rows: list = self.logical_slot_peek_changes(
+        # minimize the tmp file disk usage when calling
+        # PG_LOGICAL_SLOT_PEEK_CHANGES and PG_LOGICAL_SLOT_GET_CHANGES
+        # by limiting to a smaller batch size.
+        offset: int = 0
+        total: int = 0
+        limit: int = LOGICAL_SLOT_CHUNK_SIZE
+        count: int = self.logical_slot_count_changes(
             self.__name,
             txmin=txmin,
             txmax=txmax,
             upto_nchanges=None,
         )
+        while True:
+            changes: int = self.logical_slot_peek_changes(
+                self.__name,
+                txmin=txmin,
+                txmax=txmax,
+                upto_nchanges=None,
+                limit=limit,
+                offset=offset,
+            )
+            if not changes or total > count:
+                break
 
-        rows: list = rows or []
-        payloads: list = []
-        _rows: list = []
+            rows: list = []
+            for row in changes:
+                if re.search(r"^BEGIN", row.data) or re.search(
+                    r"^COMMIT", row.data
+                ):
+                    continue
+                rows.append(row)
 
-        for row in rows:
-            if re.search(r"^BEGIN", row.data) or re.search(
-                r"^COMMIT", row.data
-            ):
-                continue
-            _rows.append(row)
-
-        for i, row in enumerate(_rows):
-
-            logger.debug(f"txid: {row.xid}")
-            logger.debug(f"data: {row.data}")
-            # TODO: optimize this so we are not parsing the same row twice
-            try:
-                payload = self.parse_logical_slot(row.data)
-            except Exception as e:
-                logger.exception(
-                    f"Error parsing row: {e}\nRow data: {row.data}"
-                )
-                raise
-            payloads.append(payload)
-
-            j: int = i + 1
-            if j < len(_rows):
+            payloads: list = []
+            for i, row in enumerate(rows):
+                logger.debug(f"txid: {row.xid}")
+                logger.debug(f"data: {row.data}")
+                # TODO: optimize this so we are not parsing the same row twice
                 try:
-                    payload2 = self.parse_logical_slot(_rows[j].data)
+                    payload = self.parse_logical_slot(row.data)
                 except Exception as e:
                     logger.exception(
-                        f"Error parsing row: {e}\nRow data: {_rows[j].data}"
+                        f"Error parsing row: {e}\nRow data: {row.data}"
                     )
                     raise
+                payloads.append(payload)
 
-                if (
-                    payload["tg_op"] != payload2["tg_op"]
-                    or payload["table"] != payload2["table"]
-                ):
+                j: int = i + 1
+                if j < len(rows):
+                    try:
+                        payload2 = self.parse_logical_slot(rows[j].data)
+                    except Exception as e:
+                        logger.exception(
+                            f"Error parsing row: {e}\nRow data: {rows[j].data}"
+                        )
+                        raise
+
+                    if (
+                        payload["tg_op"] != payload2["tg_op"]
+                        or payload["table"] != payload2["table"]
+                    ):
+                        self.es.bulk(self.index, self._payloads(payloads))
+                        payloads: list = []
+                elif j == len(rows):
                     self.es.bulk(self.index, self._payloads(payloads))
                     payloads: list = []
-            elif j == len(_rows):
-                self.es.bulk(self.index, self._payloads(payloads))
-                payloads: list = []
-
-        if rows:
             self.logical_slot_get_changes(
                 self.__name,
                 txmin=txmin,
                 txmax=txmax,
-                upto_nchanges=len(rows),
+                upto_nchanges=None,
+                limit=limit,
+                offset=offset,
             )
+            offset += limit
+            total += len(changes)
             self.count["xlog"] += len(rows)
 
     def _payload_data(self, payload: dict) -> dict:
@@ -505,7 +560,7 @@ class Sync(Base):
                     and old_values != new_values
                 ):
                     doc: dict = {
-                        "_id": self.get_doc_id(old_values),
+                        "_id": self.get_doc_id(old_values, root.table),
                         "_index": self.index,
                         "_op_type": "delete",
                     }
@@ -607,7 +662,7 @@ class Sync(Base):
                     payload_data[key] for key in root.model.primary_keys
                 ]
                 doc: dict = {
-                    "_id": self.get_doc_id(root_primary_values),
+                    "_id": self.get_doc_id(root_primary_values, root.table),
                     "_index": self.index,
                     "_op_type": "delete",
                 }
@@ -617,7 +672,16 @@ class Sync(Base):
                     doc["_type"] = "_doc"
                 docs.append(doc)
             if docs:
-                self.es.bulk(self.index, docs)
+                raise_on_exception: Optional[bool] = (
+                    False if USE_ASYNC else None
+                )
+                raise_on_error: Optional[bool] = False if USE_ASYNC else None
+                self.es.bulk(
+                    self.index,
+                    docs,
+                    raise_on_exception=raise_on_exception,
+                    raise_on_error=raise_on_error,
+                )
 
         else:
 
@@ -801,8 +865,8 @@ class Sync(Base):
         """
         if filters.get(node.table):
             _filters: list = []
-            keys: Set = set([])
-            values: Set = set([])
+            keys: Set = set()
+            values: Set = set()
             for _filter in filters.get(node.table):
                 where: list = []
                 for key, value in _filter.items():
@@ -929,12 +993,12 @@ class Sync(Base):
 
                 if self.verbose:
                     print(f"{(i+1)})")
-                    print(f"Pkeys: {primary_keys}")
+                    print(f"pkeys: {primary_keys}")
                     pprint.pprint(row)
                     print("-" * 10)
 
                 doc: dict = {
-                    "_id": self.get_doc_id(primary_keys),
+                    "_id": self.get_doc_id(primary_keys, node.table),
                     "_index": self.index,
                     "_source": row,
                 }
@@ -972,7 +1036,7 @@ class Sync(Base):
         self._checkpoint: int = value
 
     @threaded
-    @exit_handler
+    @exception
     def poll_redis(self) -> None:
         """Consumer which polls Redis continuously."""
         while True:
@@ -980,11 +1044,24 @@ class Sync(Base):
             if payloads:
                 logger.debug(f"poll_redis: {payloads}")
                 self.count["redis"] += len(payloads)
+                self.refresh_views()
                 self.on_publish(payloads)
             time.sleep(REDIS_POLL_INTERVAL)
 
+    @exception
+    async def async_poll_redis(self) -> None:
+        """Consumer which polls Redis continuously."""
+        while True:
+            payloads: dict = self.redis.bulk_pop()
+            if payloads:
+                logger.debug(f"poll_redis: {payloads}")
+                self.count["redis"] += len(payloads)
+                await self.async_refresh_views()
+                await self.async_on_publish(payloads)
+            await asyncio.sleep(REDIS_POLL_INTERVAL)
+
     @threaded
-    @exit_handler
+    @exception
     def poll_db(self) -> None:
         """
         Producer which polls Postgres continuously.
@@ -994,9 +1071,10 @@ class Sync(Base):
         conn = self.engine.connect().connection
         conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
         cursor = conn.cursor()
-        channel: str = self.database
-        cursor.execute(f'LISTEN "{channel}"')
-        logger.debug(f'Listening for notifications on channel "{channel}"')
+        cursor.execute(f'LISTEN "{self.database}"')
+        logger.debug(
+            f'Listening for notifications on channel "{self.database}"'
+        )
         items: list = []
 
         while True:
@@ -1019,13 +1097,52 @@ class Sync(Base):
                     self.redis.bulk_push(items)
                     items = []
                 notification: AnyStr = conn.notifies.pop(0)
-                if notification.channel == channel:
+                if notification.channel == self.database:
                     payload = json.loads(notification.payload)
                     items.append(payload)
                     logger.debug(f"on_notify: {payload}")
                     self.count["db"] += 1
 
+    @exception
+    def _poll_db(self) -> None:
+        """
+        Producer which polls Postgres continuously.
+
+        Receive a notification message from the channel we are listening on
+        """
+        try:
+            self.conn.poll()
+        except OperationalError as e:
+            logger.fatal(f"OperationalError: {e}")
+            os._exit(-1)
+
+        while self.conn.notifies:
+            notification: AnyStr = self.conn.notifies.pop(0)
+            if notification.channel == self.database:
+                payload = json.loads(notification.payload)
+                self.redis.bulk_push([payload])
+                logger.debug(f"on_notify: {payload}")
+                self.count["db"] += 1
+
+    def refresh_views(self) -> None:
+        self._refresh_views()
+
+    async def async_refresh_views(self) -> None:
+        self._refresh_views()
+
+    def _refresh_views(self) -> None:
+        for node in self.root.traverse_breadth_first():
+            if node.table in self.views(node.schema):
+                if node.materialized:
+                    self.refresh_view(node.table, node.schema)
+
     def on_publish(self, payloads: list) -> None:
+        self._on_publish(payloads)
+
+    async def async_on_publish(self, payloads: list) -> None:
+        self._on_publish(payloads)
+
+    def _on_publish(self, payloads: list) -> None:
         """
         Redis publish event handler.
 
@@ -1033,6 +1150,13 @@ class Sync(Base):
         It is called when an event is received from Redis.
         Deserialize the payload from Redis and sync to Elasticsearch.
         """
+        # this is used for the views.
+        # we substitute the views for the base table here
+        for i, payload in enumerate(payloads):
+            for node in self.root.traverse_breadth_first():
+                if payload["table"] in node.base_tables:
+                    payloads[i]["table"] = node.table
+
         logger.debug(f"on_publish len {len(payloads)}")
         # Safe inserts are insert operations that can be performed in any order
         # Optimize the safe INSERTS
@@ -1078,37 +1202,55 @@ class Sync(Base):
         logger.debug(f"pull txmin: {txmin} - txmax: {txmax}")
         # forward pass sync
         self.es.bulk(self.index, self.sync(txmin=txmin, txmax=txmax))
-        self.checkpoint: int = txmax or self.txid_current
         # now sync up to txmax to capture everything we may have missed
         self.logical_slot_changes(txmin=txmin, txmax=txmax)
+        self.checkpoint: int = txmax or self.txid_current
         self._truncate: bool = True
 
     @threaded
-    @exit_handler
+    @exception
     def truncate_slots(self) -> None:
         """Truncate the logical replication slot."""
         while True:
-            if self._truncate:
-                logger.debug(f"Truncating replication slot: {self.__name}")
-                self.logical_slot_get_changes(self.__name, upto_nchanges=None)
+            self._truncate_slots()
             time.sleep(REPLICATION_SLOT_CLEANUP_INTERVAL)
 
-    @threaded
-    @exit_handler
-    def status(self):
+    @exception
+    async def async_truncate_slots(self) -> None:
         while True:
-            sys.stdout.write(
-                f"Syncing {self.database} "
-                f"Xlog: [{self.count['xlog']:,}] => "
-                f"Db: [{self.count['db']:,}] => "
-                f"Redis: [total = {self.count['redis']:,} "
-                f"pending = {self.redis.qsize:,}] => "
-                f"Elastic: [{self.es.doc_count:,}] ...\n"
-            )
-            sys.stdout.flush()
+            self._truncate_slots()
+            await asyncio.sleep(REPLICATION_SLOT_CLEANUP_INTERVAL)
+
+    def _truncate_slots(self) -> None:
+        if self._truncate:
+            logger.debug(f"Truncating replication slot: {self.__name}")
+            self.logical_slot_get_changes(self.__name, upto_nchanges=None)
+
+    @threaded
+    @exception
+    def status(self) -> None:
+        while True:
+            self._status(label="Sync")
             time.sleep(LOG_INTERVAL)
 
-    def receive(self, nthreads_polldb=None) -> None:
+    @exception
+    async def async_status(self) -> None:
+        while True:
+            self._status(label="Async")
+            await asyncio.sleep(LOG_INTERVAL)
+
+    def _status(self, label: str) -> None:
+        sys.stdout.write(
+            f"{label} {self.database} "
+            f"Xlog: [{self.count['xlog']:,}] => "
+            f"Db: [{self.count['db']:,}] => "
+            f"Redis: [total = {self.count['redis']:,} "
+            f"pending = {self.redis.qsize:,}] => "
+            f"Elastic: [{self.es.doc_count:,}] ...\n"
+        )
+        sys.stdout.flush()
+
+    def receive(self, nthreads_polldb: int) -> None:
         """
         Receive events from db.
 
@@ -1118,24 +1260,35 @@ class Sync(Base):
         2. Pull everything so far and also replay replication logs.
         3. Consume all changes from Redis.
         """
-        # start a background worker producer thread to poll the db and populate
-        # the Redis cache
-        nthreads_polldb = nthreads_polldb or NTHREADS_POLLDB
-        for _ in range(nthreads_polldb):
-            self.poll_db()
+        if USE_ASYNC:
+            self._conn = self.engine.connect().connection
+            self._conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
+            cursor = self.conn.cursor()
+            cursor.execute(f'LISTEN "{self.database}"')
+            event_loop = asyncio.get_event_loop()
+            event_loop.add_reader(self.conn, self._poll_db)
+            tasks: list = [
+                event_loop.create_task(self.async_poll_redis()),
+                event_loop.create_task(self.async_truncate_slots()),
+                event_loop.create_task(self.async_status()),
+            ]
+            event_loop.run_until_complete(asyncio.wait(tasks))
+            event_loop.close()
 
-        # sync up to current transaction_id
-        self.pull()
-
-        # start a background worker consumer thread to
-        # poll Redis and populate Elasticsearch
-        self.poll_redis()
-
-        # start a background worker thread to cleanup the replication slot
-        self.truncate_slots()
-
-        # start a background worker thread to show status
-        self.status()
+        else:
+            # start a background worker producer thread to poll the db and
+            # populate the Redis cache
+            for _ in range(nthreads_polldb):
+                self.poll_db()
+            # sync up to current transaction_id
+            self.pull()
+            # start a background worker consumer thread to
+            # poll Redis and populate Elasticsearch
+            self.poll_redis()
+            # start a background worker thread to cleanup the replication slot
+            self.truncate_slots()
+            # start a background worker thread to show status
+            self.status()
 
 
 @click.command()
@@ -1235,18 +1388,18 @@ def main(
 
     config: str = get_config(config)
 
-    show_settings(config, **kwargs)
-
-    if analyze:
-        for document in json.load(open(config)):
-            sync: Sync = Sync(document, verbose=verbose, **kwargs)
-            sync.analyze()
-        return
+    show_settings(config)
 
     with Timer():
         for document in json.load(open(config)):
             sync: Sync = Sync(document, verbose=verbose, **kwargs)
+
+            if analyze:
+                sync.analyze()
+                continue
+
             sync.pull()
+
             if daemon:
                 sync.receive(nthreads_polldb)
 
