@@ -11,6 +11,7 @@ from .settings import (
     ELASTICSEARCH_PORT,
     ELASTICSEARCH_SCHEME,
     ELASTICSEARCH_USER,
+    PG_DRIVER,
     PG_HOST,
     PG_PASSWORD,
     PG_PORT,
@@ -50,9 +51,9 @@ def get_elasticsearch_url(
         or password
         or ELASTICSEARCH_PASSWORD
     )
-    if user:
+    if user and password:
         return f"{scheme}://{user}:{quote_plus(password)}@{host}:{port}"
-    logger.debug("Connecting to Elasticsearch without authentication.")
+    logger.debug("Connecting to Elasticsearch without password.")
     return f"{scheme}://{host}:{port}"
 
 
@@ -62,18 +63,18 @@ def get_postgres_url(
     host: Optional[str] = None,
     password: Optional[str] = None,
     port: Optional[int] = None,
+    driver: Optional[str] = None,
 ) -> str:
     """Return the URL to connect to Postgres."""
     user: str = user or PG_USER
     host: str = host or PG_HOST
     password: str = _get_auth("PG_PASSWORD") or password or PG_PASSWORD
     port: str = port or PG_PORT
+    driver: str = driver or PG_DRIVER
     if not password:
         logger.debug("Connecting to Postgres without password.")
-        return f"postgresql://{user}@{host}:{port}/{database}"
-    return (
-        f"postgresql://{user}:{quote_plus(password)}@{host}:{port}/{database}"
-    )
+        return f"postgresql+{driver}://{user}@{host}:{port}/{database}"
+    return f"postgresql+{driver}://{user}:{quote_plus(password)}@{host}:{port}/{database}"
 
 
 def get_redis_url(
