@@ -5,6 +5,7 @@ import psycopg2
 import pytest
 
 from pgsync.base import subtransactions
+from pgsync.settings import NTHREADS_POLLDB
 from pgsync.sync import Sync
 
 from .helpers.utils import assert_resync_empty, noop, search
@@ -375,7 +376,7 @@ class TestNestedChildren(object):
         }
 
     def test_sync(self, sync, nodes, data):
-        """test regular sync produces the correct result."""
+        """Test regular sync produces the correct result."""
         sync.nodes = nodes
         docs = [doc for doc in sync.sync()]
         assert len(docs) == 3
@@ -972,7 +973,7 @@ class TestNestedChildren(object):
                             "pgsync.sync.Sync.status",
                             side_effect=noop,
                         ):
-                            sync.receive()
+                            sync.receive(NTHREADS_POLLDB)
                             sync.es.refresh("testdb")
 
         txmin = sync.checkpoint
@@ -1136,7 +1137,7 @@ class TestNestedChildren(object):
         continent_cls,
         book_author_cls,
     ):
-        """insert a new through child with op."""
+        """Insert a new through child with op."""
         book_author = book_author_cls(
             id=7,
             book_isbn="abc",
@@ -1800,7 +1801,7 @@ class TestNestedChildren(object):
         country_cls,
         continent_cls,
     ):
-        """insert a new non-through child with noop."""
+        """Insert a new non-through child with noop."""
         city = city_cls(
             id=5,
             name="Moscow",
@@ -2019,7 +2020,7 @@ class TestNestedChildren(object):
         country_cls,
         continent_cls,
     ):
-        """insert a new deep nested non-through child with noop."""
+        """Insert a new deep nested non-through child with noop."""
         country = country_cls(
             id=5,
             name="Marioworld",
@@ -2060,7 +2061,7 @@ class TestNestedChildren(object):
                             "pgsync.sync.Sync.status",
                             side_effect=noop,
                         ):
-                            sync.receive()
+                            sync.receive(NTHREADS_POLLDB)
                             sync.es.refresh("testdb")
 
         txmin = sync.checkpoint
@@ -2079,7 +2080,7 @@ class TestNestedChildren(object):
         country_cls,
         continent_cls,
     ):
-        """insert a new deep nested non-through fk child with op."""
+        """Insert a new deep nested non-through fk child with op."""
         nodes["children"].append(
             {
                 "table": "book_rating",
