@@ -13,6 +13,7 @@ from pgsync.exc import (
     RelationshipVariantError,
 )
 from pgsync.node import Tree
+from pgsync.settings import NTHREADS_POLLDB
 from pgsync.sync import Sync
 
 from .helpers.utils import assert_resync_empty, noop, search
@@ -326,7 +327,7 @@ class TestParentSingleChildFkOnChild(object):
         assert_resync_empty(sync, nodes)
 
     def test_null_label(self, sync, data):
-        """null label should revert back to the table name"""
+        """Null label should revert back to the table name."""
         nodes = {
             "table": "book",
             "columns": ["isbn", "title", "description"],
@@ -455,6 +456,8 @@ class TestParentSingleChildFkOnChild(object):
             "description": "Rabbits on the run",
             "isbn": "ghi",
             "publisher_id": None,
+            "seller_id": None,
+            "buyer_id": None,
             "rating": {"book_isbn": "ghi", "id": 3, "value": 3.3},
             "title": "The Rabbit Club",
         }
@@ -570,9 +573,7 @@ class TestParentSingleChildFkOnChild(object):
     def test_update_primary_key_non_concurrent(
         self, data, book_cls, rating_cls, engine
     ):
-        """
-        Test sync updates primary_key and then sync in non-concurrent mode.
-        """
+        """Test sync updates primary_key then sync in non-concurrent mode."""
         document = {
             "index": "testdb",
             "nodes": {
@@ -754,7 +755,7 @@ class TestParentSingleChildFkOnChild(object):
                             "pgsync.sync.Sync.status",
                             side_effect=noop,
                         ):
-                            sync.receive()
+                            sync.receive(NTHREADS_POLLDB)
                             sync.es.refresh("testdb")
 
         docs = search(sync.es, "testdb")
@@ -1054,7 +1055,7 @@ class TestParentSingleChildFkOnChild(object):
                             "pgsync.sync.Sync.status",
                             side_effect=noop,
                         ):
-                            sync.receive()
+                            sync.receive(NTHREADS_POLLDB)
                             sync.es.refresh("testdb")
 
         docs = search(sync.es, "testdb")
@@ -1173,7 +1174,7 @@ class TestParentSingleChildFkOnChild(object):
                             "pgsync.sync.Sync.status",
                             side_effect=noop,
                         ):
-                            sync.receive()
+                            sync.receive(NTHREADS_POLLDB)
                             sync.es.refresh("testdb")
 
         docs = search(sync.es, "testdb")
